%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%   $RCSfile: ale_redefinitions.pl,v $
%%  $Revision: 1.17 $
%%      $Date: 2003/10/10 16:36:03 $
%%     Author: Stefan Mueller (Stefan.Mueller@uni-bremen.de)
%%    Purpose: redefinitions of some ALE stuff
%%             needed to have pointers to solutions
%%   Language: Prolog
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Log: ale_redefinitions.pl,v $
%% Revision 1.17  2003/10/10 16:36:03  stefan
%% fixed a backtracking bug
%%
%% Revision 1.16  2003/10/08 20:40:33  stefan
%% reverse_count_lex_check is now back to ale_redefinitions since it is
%% needed in both [incr TSDB()] and in the chart_display interactive level.
%%
%% ale.pl causes an exception which is too strong. It breaks the interactive level.
%% This code records all missing words, produces a warning and fails.
%%
%% Revision 1.15  2003/08/13 12:06:31  stefan
%% fixed edge_limit check and exception catch in trale-itsdb and ale_redefinitions check and exception catch in trale-itsdb and ale_redefinition
%%
%% Revision 1.14  2003/08/09 15:53:07  dm
%% aufgeraeumt und modularisiert
%%
%% Revision 1.13  2003/08/09 14:42:41  dm
%% Removed temporary redefinition of reverse_count_lex_check/5. We now
%% have the redefinition from trale_patches.pl as part of the
%% redefinitions that are used for itsdb as defined in trale-itsdb.pl
%%
%% Revision 1.12  2003/08/06 10:08:05  stefan
%% reverse_... now fails if words are missing
%%
%% Revision 1.11  2003/08/06 10:04:58  stefan
%% reverse_count_lex_check as modified by Gerald did not work
%% for tsdb since I need all missing words. If Gerald insists
%% on throwing exceptions he should throw an exception that
%% informs the user of all missing words.
%%
%% Revision 1.10  2003/06/11 14:40:45  stefan
%% reverse_count -> reverse_count_lex_check
%%
%% Revision 1.9  2003/05/22 09:25:53  stefan
%% fixed multiple error messages in case we reach the edge_limit
%%
%% Revision 1.8  2003/05/19 15:39:21  stefan
%% Added exception handling in case of memory overflow. Error information
%% is asserted that is used by [incr TSDB()].
%%
%% Revision 1.7  2003/05/13 20:06:47  stefan
%% put init_chart_display before reverse_count because otherwise
%% we may have an old chart display open that displays invalid edges
%% after an unsuccessful reverse_count (missing lexical entries).
%%
%% Revision 1.6  2003/05/13 09:10:49  stefan
%% included a (wrong) fix to rule/1 in ale_redefinitions.pl
%%
%% Revision 1.5  2003/05/09 09:21:00  stefan
%% *** empty log message ***
%%
%% Revision 1.4  2003/05/04 17:20:37  stefan
%% removed all_words_known, since the check is done in reverse_count now
%%
%% Revision 1.3  2003/05/02 17:27:11  dm
%% Added init_chart_display_if_on to rec so that we get chart display
%% output with ordinary rec calls
%%
%% Revision 1.2  2003/05/02 16:36:36  dm
%% - eliminated redefinition of rect and get_and_show_solutions
%%   since test_suite_handling.pl now integrates what they did
%% - eliminated all_words_known/1 and error/2 code since
%%   trale_patches.pl now contains a redefined version of
%%   reverse_count taking care of early lookup of words
%%
%% Revision 1.1  2003/05/02 02:07:12  dm
%% added Stefan's chart_display code
%%
%%
%% Revision 1.1  2003/04/14 10:06:13  stefan
%% Initial revision
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic rule_counter/1.

:- assert(rule_counter(-1)).

% --------------------------------

:- abolish(clear/0).

clear :-
  retractall(to_rebuild(_)),
  retractall(edge(_,_,_,_,_,_,_)),
  retractall(parsing(_)),
  retractall(error(_,_)),
  retractall(solution(_)),      % solutions for the chart display   % ADDED St. M.
  retractall(num(_)), % edge index
  retractall(go(_)).  % interpreter go flag

% -----------------------------------------------------------------------------
% rec(+Ws:<word>s, Tag:<var_tag>, SVs:<svs>, Iqs:<ineq>s)
% -----------------------------------------------------------------------------
% Ws can be parsed as category Tag-SVs with inequations Iqs;  Tag-SVs 
%  uninstantiated to start
% -----------------------------------------------------------------------------

:- abolish(rec/4).

rec(Ws,Tag,SVs,Residue) :-
  clear,
  assert(parsing(Ws)),
  asserta(num(0)),

  init_chart_display_if_on, % ADDED

  ( current_predicate(lex,lex(_,_))
  -> reverse_count_lex_check(Ws,[],WsRev,0,Length),
     CLength is Length - 1,
  
     % Register empty elements for chart display  % ADDED
     (  current_predicate(draw_empty_cats/1)      % ADDED
     -> draw_empty_cats(Length)	                  % ADDED
     ;  true ),                                   % ADDED

     functor(Chart,chart,CLength),
     build(WsRev,Length,Chart),
     retract(to_rebuild(Index)),
     call_residue((clause(edge(Index,0,Length,Tag,SVs,_,_),true),
                   extensionalise(Tag,SVs)),Residue),
     assert(solution(Index))
  ;  raise_exception(ale(no_lex))
  ).

% -----------------------------------------------------------------------------
% rec(+Ws:<word>s, Tag:<var_tag>, SVs:<svs>, Iqs:<ineq>s, ?Desc:<desc>)
% -----------------------------------------------------------------------------
% Like rec/3, but Tag-SVs also satisfies description, Desc.
% -----------------------------------------------------------------------------

:- abolish(rec/5).

rec(Ws,TagOut,SVsOut,Desc,Residue) :-
  clear,
  assert(parsing(Ws)),
  asserta(num(0)),

  init_chart_display_if_on, % ADDED

  (  current_predicate(lex,lex(_,_))
  -> reverse_count_lex_check(Ws,[],WsRev,0,Length),
     CLength is Length - 1,

     % Register empty elements for chart display  % ADDED
     (  current_predicate(draw_empty_cats/1)      % ADDED
     -> draw_empty_cats(Length)	                  % ADDED
     ;  true ),                                   % ADDED

     functor(Chart,chart,CLength),
     build(WsRev,Length,Chart),
     retract(to_rebuild(Index)),
     call_residue((clause(edge(Index,0,Length,Tag,SVs,_,_),true),
                   (secret_noadderrs
                   ; secret_adderrs,
                     fail),
                   add_to(Desc,Tag,SVs),
                   deref(Tag,SVs,TagOut,SVsOut),
                   extensionalise(TagOut,SVsOut),
                   (secret_adderrs
                   ; secret_noadderrs,
                     fail)),Residue),
     assert(solution(Index))
  ; raise_exception(ale(no_lex))
  ).

% -----------------------------------------------------------------------------

:- abolish(edge_assert/7).

edge_assert(Left,Right,TagOut,SVsOut,Dtrs,RuleName,N) :- 
  no_interpreter,                                      
  !,gennum(N),
  asserta(to_rebuild(N)),
  asserta(edge(N,Left,Right,TagOut,SVsOut,Dtrs,RuleName)),
  draw_edge(RuleName,N,Left,Right,Dtrs).	% ADDED
%  format('Edge added: Number: ~w, Left: ~w, Right: ~w, Rule: ~w~n',
%         [N,Left,Right,RuleName]), % DEBUG
%  ttyflush.

edge_assert(Left,Right,TagOut,SVsOut,Dtrs,RuleName,N) :-
  !,nl,
  length(Dtrs,ND),
  (show_res -> frozen_term([TagOut|SVsOut],Frozen) ; Frozen = []),
  ( print_edge(Left,Right,TagOut,SVsOut,Dtrs,RuleName,ND,Frozen)
  -> gennum(N),
     asserta(to_rebuild(N)),
     asserta(edge(N,Left,Right,TagOut,SVsOut,Dtrs,RuleName)),
     draw_edge(RuleName,N,Left,Right,Dtrs)).	% ADDED

draw_edge(RuleName,N,Left,Right,Dtrs) :-
  (current_predicate(draw_item/5) ->
   (( RuleName = lexicon ->
     get_words_from_parse_string(Left,Right,[NameN])   
   ; otherwise ->
     NameN = RuleName
   ),
   draw_item(N,Left,Right,NameN,Dtrs))
  ;true).

/*

It's better to do the memory check in an exception handler and call
the edgelimit test in gennum directly.

ts_error(Type):-
	( clause(error(_,_), true) -> % There is already an error
		 fail
	; Type == edgecount ->
		 ((current_predicate(edge_limit/1),
                     edge_limit(Max)) -> 
                     num(N),
                     ( N < Max -> true
                     ; assertz(error(edge_limit,'Maximum edge count')),
                       format(user_error,'~N**ERROR: Edge limit ~w reached.~N',[Max]),
                         fail
                     )
                 ; true
                 )
	; Type == memory ->
		 statistics(memory,[Mem,_]),
		 ( Mem < 253000000 -> true % SICStus Prolog maximum = 256000000 (not linux!)
		 ; trimcore,
			assertz(error(memory_limit,'Maximum memory usage')),
                        format(user_error,'~N**ERROR: Maximum memory usage!~N'),
			fail
		 )
	; assertz(error(unknown,'Unkown error')),
	  fail
	).

*/

%---------------------------------------------------------------------
% Substitutes gennum/1 to take maximum number of edges into account.

:- abolish(gennum/1).

gennum(N) :-
  retract(num(N)),
  NewN is N+1,
  asserta(num(NewN)),
  ((current_predicate(edge_limit/1),
    edge_limit(Max),
    Max < NewN) ->
    ( error(edge_limit,_) -> 
      fail  % print the message only once
    ; assertz(error(edge_limit,'Maximum edge count')),
      format(user_error,'~n* ERROR: Edge limit ~w reached.~n~n',[Max]),
      fail
    )
 ;  true
 ).


/*:- abolish(rule/1).

rule(RuleName):-
  call_residue(clause(alec_rule(RuleName,DtrsDesc,_,Moth,_,_),true),_),
%  (RuleName rule Moth ===> DtrsDesc),
  empty_assoc(AssocIn),  
  call_residue((satisfy_dtrs(DtrsDesc,DtrCats,[],Dtrs,gdone),
                (secret_noadderrs
                ; secret_adderrs,
                  fail),
		add_to(Moth,TagMoth,bot),
                CatsOut = [TagMoth-bot|DtrCats],
		extensionalise_list(CatsOut)),Residue),
  \+ \+ (((current_predicate(portray_rule,portray_rule(_,_,_,_)),
	   portray_rule(TagMoth,bot,Dtrs,Residue)) -> true
	 ;
           nl, write('RULE: '), write(RuleName),
           build_iqs(Residue,Iqs,FSResidue),
	   (show_res -> residue_args(FSResidue,ResArgs,CatsOut) ; ResArgs = CatsOut),
           duplicates_list(ResArgs,AssocIn,DupsMid,AssocIn,VisMid,0,NumMid),
  	   duplicates_iqs(Iqs,DupsMid,DupsMid2,VisMid,_,NumMid,_),
           nl, nl, write('MOTHER: '), nl,
           nl, tab(2), pp_fs(TagMoth,bot,DupsMid2,DupsMid3,AssocIn,VisMid2,2,AssocIn,HDMid),
           nl, nl, write('DAUGHTERS/GOALS: '),   
           show_rule_dtrs(Dtrs,DupsMid3,DupsMid4,VisMid2,VisMid3,HDMid,HDMid2),
	   nl,nl, tab(2), pp_iqs(Iqs,DupsMid4,DupsOut,VisMid3,VisOut,2,HDMid2,HDOut),
  	   ((show_res,FSResidue \== [])
  	   -> nl, nl, write('Residue:'), pp_residue(FSResidue,DupsOut,_,VisOut,_,HDOut,_)
           ; true), nl
	 ),
         query_proceed).
*/


% -------------------------------------------
% add unknown word lookup -> fails if unknown words were detected
%                            reports all missing words, asserts them for [incr TSDB()]

:- dynamic error/2.

:- abolish([reverse_count_lex_check/5]).

:- use_module(library(lists),[memberchk/2]).

reverse_count_lex_check(ListIn,Acc,ListOut,CountIn,CountOut) :-
   retractall(error(word,_)),
   reverse_count_lex_check(ListIn,Acc,ListOut,CountIn,CountOut,[],Errors),
   ( Errors==[] -> true
   ; otherwise -> format(user_error,'~n* ERROR: The following words are unknown: ',[]),
     list_to_atom(Errors,32,ErrorA),
     format(user_error,'~w~n~n',[ErrorA]),
     fail
   ).           % This fails if there are errors,
                % instead of raising an exception
                % with raise_exception(ale(unk_word(Errors))) since
                % processing with itsdb should continue even if words 
                % are missing
   
reverse_count_lex_check([],Xs,Xs,Count,Count,Errors,Errors).
reverse_count_lex_check([X|Xs],Ys,Zs,CountIn,Count,ErrorsIn,ErrorsOut):-
  (  lex(X,_)
  -> ErrorsOut=ErrorsMid
  ;  (  memberchk(X,Ys)
     -> ErrorsOut=ErrorsMid
     ;  ErrorsOut=[X|ErrorsMid],
        assert(error(word,X))
     )
  ),
  CountInPlus1 is CountIn+1,
  reverse_count_lex_check(Xs,[X|Ys],Zs,CountInPlus1,Count,ErrorsIn,ErrorsMid).

list_to_string([Last],_,LastS) :-
  name(Last,LastS).

list_to_string([H|T],D,String) :-
  list_to_string(T,D,S1),
  name(H,HS),
  append(HS,[D|S1],String).

list_to_atom(List,Delimiter,Atom) :-
  list_to_string(List,Delimiter,String),
  name(Atom,String).
