%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%   $RCSfile: statistics.pl,v $
%%  $Revision: 1.2 $
%%      $Date: 2003/05/27 02:43:09 $
%%     Author: Stefan Mueller (Stefan.Mueller@uni-bremen.de)
%%    Purpose: Output of TSDB++ statistics, communication
%%             with capi interface, see trale-tsdb.c
%%   Language: Prolog
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Log: statistics.pl,v $
%% Revision 1.2  2003/05/27 02:43:09  dm
%% Changed lex/3 to lex/2 as it is now defined in ale.pl and removed call
%% to extensionalise in c_words(0), which as far as I see is not needed there.
%%
%% Revision 1.1  2003/05/04 23:04:38  dm
%% Next revision of files by Stefan.
%%
%%
%% Initial revision
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

c_words(0) :-
        retractall(counter(_)),
        assert(counter(0)),
        call_residue(lex(_Word,_FS),_Residue),
        retract(counter(C)),
        C1 is C + 1,
        assert(counter(C1)),
        fail.

c_words(Counter) :-
        retract(counter(Counter)).



c_lexrules(0) :-
        retractall(counter(_)),
        assert(counter(0)),
        ( (RuleName lex_rule Desc1 **> Desc2 if Cond morphs Morphs)
        ; (RuleName lex_rule Desc1 **> Desc2 morphs Morphs),
            Cond = true
        ),
        empty_assoc(AssocIn),
        call_residue((add_to(Desc1,Tag1,bot),
                      add_to(Desc2,Tag2,bot),
                      nv_replace_body(Cond,_Goal,Args,[],AssocIn),
                      ArgsOut = [Tag1-bot,Tag2-bot|Args],
                      extensionalise_list(ArgsOut)),_Residue),
        retract(counter(C)),
        C1 is C + 1,
        assert(counter(C1)),
        fail.

c_lexrules(Counter) :-
        retract(counter(Counter)).

c_rules(0) :-
        retractall(counter(_)),
        assert(counter(0)),
        clause(alec_rule(_RuleName,DtrsDesc,_,Moth,_,_),true),
%  empty_assoc(AssocIn),  
        call_residue((satisfy_dtrs(DtrsDesc,DtrCats,[],_Dtrs,gdone),
                      (secret_noadderrs
                      ; secret_adderrs,
                          fail),
                      add_to(Moth,TagMoth,bot),
                      CatsOut = [TagMoth-bot|DtrCats],
                      extensionalise_list(CatsOut)),_Residue),
        retract(counter(C)),
        C1 is C + 1,
        assert(counter(C1)),
        fail.

c_rules(Counter) :-
        retract(counter(Counter)).


c_macros(0) :-
        retractall(counter(_)),
        assert(counter(0)),
        (MacroName macro Desc),
        call_residue((add_to(Desc,Tag,bot),
                MacroName =.. [_Name|MacroArgDescs],
                mg_sat_list(MacroArgDescs,MacroArgs),
%                MacroSat =.. [Name|MacroArgs],
                ArgsOut = [Tag-bot|MacroArgs],
		extensionalise_list(ArgsOut)),_Residue),
        retract(counter(C)),
        C1 is C + 1,
        assert(counter(C1)),
        fail.

c_macros(Counter) :-
        retract(counter(Counter)).

% definied in ale.pl  fails if there are no edges
c_edges(N) :-
        call_residue(count_edges(N),_),!.

c_edges(0).

/*
c_edges(0) :-
        retractall(counter(_)),
        assert(counter(0)),
        call_residue(clause(edge(_Index,_Start,_End,_Tag,_SVs,_Dtrs,_RuleName),true),_Residue),
        retract(counter(C)),
        C1 is C + 1,
        assert(counter(C1)),
        fail.

c_edges(Counter) :-
        retract(counter(Counter)).
*/

c_result_edges(0) :-
        retractall(counter(_)),
        assert(counter(0)),
        solution(Index),
        count_edges_index(Index,Counter),
        retract(counter(C)),
        C1 is C + Counter,
        assert(counter(C1)),
        fail.

c_result_edges(Counter) :-
        retract(counter(Counter)).

        
count_edges_index(Index,Counter) :-
        call_residue(clause(edge(Index,_Start,_End,_Tag,_SVs,Dtrs,_RuleName),true),_Residue),
        count_dtrs(Dtrs,NofDtrs),
        Counter is 1 + NofDtrs.

count_dtrs(Dtrs,NofDtrs) :-
  count_dtrs(Dtrs,0,NofDtrs).

count_dtrs([],Counter,Counter).
count_dtrs([empty(DtrId,_)|R],CounterIn,CounterOut) :-
        call_residue(empty_cat(DtrId,_L,_Tag,_SVs,Dtrs,_RuleName),_Residue),
        count_dtrs(Dtrs,EmptyDtrs),
        count_dtrs(R,CounterDtrs),
        CounterOut is CounterIn + 1 + EmptyDtrs + CounterDtrs.

count_dtrs([F|R],CounterIn,CounterOut) :-
        count_edges_index(F,Counter),
        count_dtrs(R,0,CounterDtrs),
        CounterOut is CounterIn + Counter + CounterDtrs.


        
        

c_words_in_parse(0) :-
        retractall(counter(_)),
        assert(counter(0)),
        call_residue(clause(edge(_Index,_Start,_End,_Tag,_SVs,[],lexicon),true),_Residue),
        retract(counter(C)),
        C1 is C + 1,
        assert(counter(C1)),
        fail.

c_words_in_parse(Counter) :-
        retract(counter(Counter)).

c_types(C) :-
        count(type/1,C).

