% $Id: extend_access.pl,v 1.10 2003/05/08 17:46:24 dm Exp $
%
% derived from Revision 1.13  of ConTroll/system/extend_access.pl
%

% operator for defining access relations
:- op(1125,xfx,':=='). 

:- dynamic userDefinitionError/0.
:- dynamic accessRelation/7. 

:- use_module(library(lists),[reverse/2]).

%----------------------------------------------------

% to load files with relations and rules
extend_access(FileList,OutFileName) :-
   abolish(access_rule/3),
   abolish(access_suffix/2),
   FileList,
   extend_access(OutFileName).

% to extend relations with rules, both of which are already loaded
extend_access(OutFileName):-
   checkUserDefinitions,
   retractall(accessRelation(_,_,_,_,_,_,_)),
   t_open(OutFileName,write,OutStream),
   writeFileHeader(OutStream),
   doExtension(OutStream),
   format('~N Extend access mechanism has created file "~w".~n',[OutFileName]),
   close(OutStream).

% to run the example
test_extend_access :- 
   extend_access([extend_access_example],eae_result).

%----------------------------------------------------
% predicates to make sure the specification provided by the user is ok

checkUserDefinitions :-
   retractall(userDefinitionError),
   checkAccessSuffix,
   checkAccessRule,!,
   (userDefinitionError ->  
       format(user_error,' aborting~n',[]),
       abort
   ;
       true).

checkAccessRule :-
   getAccessRule(T1,F,T2),
   checkSuffix(T1,T1,F,T2),
   checkSuffix(T2,T1,F,T2),
   fail.
checkAccessRule.

checkSuffix(T,T1,F,T2) :-
   findall(S, access_suffix(T,S), L),
   treatSuffixResult(L,T,T1,F,T2).

treatSuffixResult([],T,T1,F,T2) :-
   !,
   format(user_error,'~N**Error: No suffix defined for "~w" used in "',[T]),
   format(user_error,'access_rule(~w,~w,~w)~n',[T1,F,T2]),
   assert(userDefinitionError).

treatSuffixResult([_],_,_,_,_):-
   !.
treatSuffixResult(L,T,T1,F,T2) :-
   format(user_error,'~N**Error: Multiple suffixes ',[]),
   writeMultipleSuffixes(L),
   format(user_error,'defined for type "~w"~n',[T]),
   format(user_error,'         used in "access_rule(~w,~w,~w)"~n',[T1,F,T2]),
   assert(userDefinitionError).

getAccessRule(T1,F,T2) :-
   catch(access_rule(T1,F,T2),
	 existence_error(_,_,_,_,_),
	 (format(user_error,'~N**Error: No access_rule defined.~n',[]),
	     assert(userDefinitionError),
	     fail)).
getAccessSuffix(T,S) :-
   catch(access_suffix(T,S),
	 existence_error(_,_,_,_,_),
	 (format(user_error,'**Error: No access_suffix defined.~n',[]),
	     assert(userDefinitionError),
	     fail)).
	
writeSuffix(X) :-
   atom_codes(AtomX,X),
   write('"'),
   t_write(AtomX),
   write('"').

writeMultipleSuffixes([H|T]) :-
   writeSuffix(H),
   (\+ (T = []) ->
       write(', ')
   ;
       write(' ')),	    
   writeMultipleSuffixes(T).
writeMultipleSuffixes([]).

checkAccessSuffix :-
   findall(S,getAccessSuffix(_,S),L),
   checkAccessSuffixAux(L,_).

checkAccessSuffixAux([],_).
checkAccessSuffixAux([H|T],Error) :-
   findall((E,H),member(E,T),AllMem),
   checkEachSuffix(AllMem,Error),
   checkAccessSuffixAux(T,Error).

checkEachSuffix([],_).
checkEachSuffix([(E,H)|T],Error):-
   reverse(E,RevE),
   reverse(H,RevH),
   openEndedList(RevE,RevEOpen),
   openEndedList(RevH,RevHOpen),
   findall((RevH,RevE),append(RevHOpen,_,RevE),X1),
   analyzeSuffixResult(X1,Error),
   (RevH = RevE ->
       true
   ;
       findall((RevE,RevH),append(RevEOpen,_,RevH),X2),
       analyzeSuffixResult(X2,Error)),
   checkEachSuffix(T,Error).

openEndedList([],_).
openEndedList([H|T1],[H|T2]) :-
   openEndedList(T1,T2).

analyzeSuffixResult([],_).
analyzeSuffixResult([(RevA,RevB)|_],Error) :-
   (Error == true ->
       true
   ;
       format(user_error,'**Error: Identical or suffix-identical suffixes detected:~n',[]),
       Error = true),
   reverse(RevA,A),
   access_suffix(TA,A),!,
   reverse(RevB,B),
   access_suffix(TB,B),!,
   tab(9),
   write('access_suffix('),
   t_write(TB),
   write(','),
   writeSuffix(B),
   write(')'),
   write(' --- '),
   write('access_suffix('),
   t_write(TA),
   write(','),
   writeSuffix(A),
   write(')'),nl,
   assert(userDefinitionError).

%----------------------------------------------------
% here starts the real algorithm

doExtension(OutStream) :-
   :==(H,T),
   treatSourceRelation(H,T,OutStream),
   findall(NewRel,applyARule(H,NewRel,OutStream),NewRelL),
   treatDerivedRelations(NewRelL,OutStream),
   fail.
doExtension(_).

treatDerivedRelations([H|T],OutStream) :-
   findall(NewRel,applyARule(H,NewRel,OutStream),NewRelL),
   treatDerivedRelations(NewRelL,OutStream),
   treatDerivedRelations(T,OutStream).
treatDerivedRelations([],_).

%----------------------------------------------------

treatSourceRelation(H,T,OutStream) :-
   getTyping(H,RelName,PureRelName,Type,PureArgs,ArgTypes),
   NewH =.. [RelName|PureArgs], 
   treatNewRelation(NewH,T,RelName,Type,
		    PureRelName,ArgTypes,OutStream,true).

getTyping(H,RelName,PureRelName,Type,PureArgs,ArgTypes) :-
   H =.. [RelName|Args],
   getRelTyping(PureRelName,Type,RelName),
   getArgsTyping(Args,PureArgs,ArgTypes,RelName).

% getRelTyping(-,-,+)
getRelTyping(PureRelName,OutType,OutRelName) :-
   atom_codes(OutRelName,OutRelNameL),
   access_suffix(OutType,OutSuffix),
   append(PureRelNameL,OutSuffix,OutRelNameL),
   atom_codes(PureRelName,PureRelNameL).

getArgsTyping([],[],[],_).
getArgsTyping([H|R],[F|FR],[T|TR],RelName) :-
   compound(H),
   H = F-T,
   !,
   getArgsTyping(R,FR,TR,RelName).

% If we do NOT want to enforce typing specified on each argument comment
% in the first, else the second clause: %TRALE
getArgsTyping([F|R],[F|FR],[_|TR],RelName) :-
   getArgsTyping(R,FR,TR,RelName).

%getArgsTyping([F|R],[F|FR],[T|TR],RelName) :-
%	unconstrainedType(T),
%	write('**Error: No typing specified in relation "'),
%	t_write(RelName), write('"'), nl,
%	tab(9), write('for argument "'),
%	t_write(F),	write('".'),nl,
%	getArgsTyping(R,FR,TR,RelName).

% A fact to hold the name of the unconstrained type:
%unconstrainedType(bot).

%----------------------------------------------------

applyARule(InHead,OutHead,OutStream) :-
   accessRelation(InHead,_,InRelName,InType,PureRelName,ArgTypes,_),
   access_rule(OutType,Path,InType),
   getSuffixedName(PureRelName,OutType,OutRelName),
   getFreeArgL(InHead,ArgL),
   OutHead =.. [OutRelName | ArgL],
   RelationCall =.. [InRelName | ArgL],
   OutTail = (Path:(@RelationCall)), %TRALE
   treatNewRelation(OutHead,OutTail,OutRelName,OutType,
		    PureRelName,ArgTypes,OutStream,fail).

treatNewRelation(OutHead,OutTail,OutRelName,OutType,
                 PureRelName,ArgTypes,OutStream,First) :-
   (alreadyDefined(OutHead,AlreadyHead,AlreadyTail,OutRelName,
		   WarningType) ->
       writeWarning(OutHead,OutTail,AlreadyHead,AlreadyTail,
		    WarningType,First),
       fail
   ;
       (First -> 
	   nl(OutStream), 
	   write(OutStream,'%% ---- '),
	   t_write(OutStream,OutRelName),
	   write(OutStream,' ------------------------------- '),
	   nl(OutStream), nl(OutStream)
       ;   
	   true),
       
       getFreeArgL(OutHead,FreeArgL),
       FreeOutHead =.. [OutRelName | FreeArgL],
       assertz(accessRelation(FreeOutHead,OutTail,OutRelName,OutType,
			      PureRelName,ArgTypes,First)),
       
% no extra typing statements in TRALE:
% writeTyping(OutRelName,OutType,ArgTypes,OutStream),
%
% Instead  one should write out auxiliary macro introducing
% the typing (TODO)
       
       writeDefinition(OutHead,OutTail,OutStream),
       nl(OutStream)).

alreadyDefined(OutHead,OutHead,AlreadyTail,_,a(exact)) :-
   fail_if_undefined(macro(OutHead,AlreadyTail)),!. % TRALE

alreadyDefined(OutHead,OutHead,AlreadyTail,_,b(exact,First)) :-
   accessRelation(OutHead,AlreadyTail,_,_,_,_,First),!.

% Currently in ConTroll each relation must have a unique name, even if the
% arity is different. Once this is changed, the following clauses with a result
% argument containing 'arity' in the following can be eliminated 
% and the program will no longer enforce this restriction:
%
%alreadyDefined(_,AlreadyHead,AlreadyTail,RelName,b(arity,First)) :-
%	accessRelation(AlreadyHead,AlreadyTail,RelName,_,_,_,First),!.
%
%alreadyDefined(_,AlreadyHead,AlreadyTail,RelName,a(arity)) :-
%	:=(AlreadyHead,AlreadyTail),
%	AlreadyHead =.. [RelName|_],!.


writeWarning(OutHead,OutTail,AlreadyHead,AlreadyTail,WarningType,First) :-
   write('* WARNING: Attempt to redefine '),
   OutHead =..[OutName|_],
   write('"'),t_write(OutName),write('"'),
% functor(OutHead,OutName,OutArity),
% write(OutName), write('/'), write(OutArity),
   nl, tab(2), write('- It is already defined '),
   (WarningType=a(_) -> 
       write('as PRIMARY RELATION "')
   ;
       (WarningType=b(_,true) ->
	   write('as ACCESS RELATION  "')
       ;
	   write('by ACCESS EXPANSION "'))),
% functor(AlreadyHead,AlreadyName,AlreadyArity),
% write(AlreadyName), write('/'), write(AlreadyArity), 
   writeNumbervars(AlreadyHead),
   write(' macro '),
   writeNumbervars(AlreadyTail),	write('"'), 
   
   nl,tab(2),write('- Ignoring redefinition by access '), 
   (First ->
       write('RELATION ') 
   ;
       write('EXPANSION')),
   write(' "'),
   writeNumbervars(OutHead),
   write(' := '),
   writeNumbervars(OutTail),
   write('"'),
   nl,tab(4), 
   write('and stopping its further expansion'), 
   
   ((WarningType=a(arity) ; WarningType=b(arity,_)) ->
       nl,tab(2), 
       write('- Note: Predicate names must differ'),
       write('even for different arities !')
   ;
       true),
   
   nl,nl.


% writeNumbervars/1:  write something after numbervars 
% so that the variables only get changed for writing.  
writeNumbervars(X):- 
   numbervars(X,0,_),
   t_write(X), fail.  
writeNumbervars(_).


% getSuffixedName(+,+,-)
getSuffixedName(PureRelName,OutType,OutRelName) :-
   atom_codes(PureRelName,PureRelNameL),
   access_suffix(OutType,OutSuffix),!,
   append(PureRelNameL,OutSuffix,OutRelNameL),
   atom_codes(OutRelName,OutRelNameL).

getFreeArgL(Head,FreeArgL) :-
   Head =.. [_|ArgL],
   sameLength(ArgL,FreeArgL).

sameLength([],[]).
sameLength([_|T],[_|R]) :-
   sameLength(T,R).

%----------------------------------------------------
% Output Predicates
%----------------------------------------------------

writeFileHeader(OutStream) :-
   write(OutStream,':- multifile ale_macro/2.'),
   nl(OutStream),
   nl(OutStream),
   write(OutStream,'%% File contains access relations generated by the following recipes:'),
   nl(OutStream), write(OutStream,'%%'), nl(OutStream),
   writeAccessRuleDefs(OutStream),
   write(OutStream,'%%'), nl(OutStream),
   writeAccessSuffixDefs(OutStream),
   nl(OutStream).

writeAccessRuleDefs(OutStream) :-
   access_rule(T1,F,T2),
   write(OutStream,'%% '),
   t_write(OutStream,access_rule(T1,F,T2)),
   write(OutStream,'.'),
   nl(OutStream),
   fail.
writeAccessRuleDefs(_).

writeAccessSuffixDefs(OutStream) :-
   access_suffix(T,Suffix),
   atom_codes(Atom,Suffix),
   atom_concat(Atom,'"',Tmp),
   atom_concat('"',Tmp,OutputString),
   write(OutStream,'%% '),
   t_write(OutStream,access_suffix(T,OutputString)),
   write(OutStream,'.'),
   nl(OutStream),
   fail.
writeAccessSuffixDefs(_).

%----------------------------------------------------

writeTyping(RelName,Type,ArgTypes,OutStream) :-
   t_write(OutStream,RelName),
   writeArgsTyping(ArgTypes,OutStream),
   write(OutStream,' **> '),
   t_write(OutStream,Type),
   write(OutStream,'.'),
   nl(OutStream).

writeArgsTyping(L,OutStream) :-
   (L = [] ->
       true
   ;
       L = [H|T],
       write(OutStream,'('),
       t_write(OutStream,H),
       writeArgsTypingAux(T,OutStream),
       write(OutStream,')')).

writeArgsTypingAux([H|T],OutStream) :-
   write(OutStream,','),
   t_write(OutStream,H),
   writeArgsTypingAux(T,OutStream).
writeArgsTypingAux([],_).

writeDefinition(Head,Tail,OutStream) :-
   numbervars(Head-Tail,0,_),
   t_write(OutStream,Head),
   write(OutStream,' := '),
   t_write(OutStream,Tail),
   write(OutStream,'.'),
   nl(OutStream).
