% $Id: ghooks.pl,v 1.51 2003/10/16 04:47:37 dm Exp $
%
% This file defines a pretty printing hook for the grisu interface.
%
% Authors: Detmar Meurers and Dale Gerdemann
% 
% TODO:
%   - relational attachments
%
% ------------------------------------------------------

:- multifile portray_fs/10, portray_approps/2.

:- use_module(library(lists),[select/3]).
:- use_module(library(ugraphs),[vertices_edges_to_ugraph/3]).
:- use_module(library(assoc),[assoc_to_list/2,put_assoc/4,del_assoc/4,
                              empty_assoc/1,get_assoc/3]).

% also: t_samsort/3 from generally loaded sicstus site additions

% ------------------------------------------------------
% Load some other parts of the system this builds on:

% Feature ordering
:- ensure_loaded(trale_home(feature_ordering)).

% Socket support:
:- ensure_loaded(trale_home(socket_interface)).

% Unfilling code:
:- ensure_loaded(trale_home(unfill)).

% should always be loaded before other hooks are:
:- [trale_home(save_results)].

% ------------------------------------------------------
% switches:

:- dynamic grisu_flag/0, grisu_debug_flag/0.

grisu_on :-
   retractall(grisu_flag),
   assert(grisu_flag).

grisu_off :- 
   retractall(grisu_flag).

:- grisu_on.

% --------------------------

grisu_debug :- % shows the output normally sent to grisu on standard out
   retractall(grisu_debug_flag),
   assert(grisu_debug_flag).

grisu_nodebug :-
   retractall(grisu_debug_flag).

% --------------------------

:- dynamic pref_struc_flag/1.

pref_struc_on :-
   asserta(pref_struc_flag(true)).

pref_struc_off :-
   asserta(pref_struc_flag(fail)).

pref_struc_restore :-  % restore to previous value
   retract(pref_struc_flag(_)).

pref_struc_flag :-     % 0-arity pred for testing the value of this flag
   pref_struc_flag(X), !,
   call(X).

% ------------------------------------------------------

grisu_pp_fs(FS,FlagAss) :-
   \+ \+ (deref(FS,Ref,SVs),
          grisu_pp_fs_act(Ref,SVs,FlagAss)).
grisu_pp_fs(Ref0,SVs0,FlagAss) :-
   \+ \+ (deref(Ref0,SVs0,Ref1,SVs1),
          grisu_pp_fs_act(Ref1,SVs1,FlagAss)).

grisu_pp_fs_act(Ref,SVs,FlagAss) :-
   empty_assoc(Empty),
   duplicates(Ref,SVs,Empty,DupsIn,Empty,_,0,_),
   grisu_pp_fs(Ref,SVs,DupsIn,_,Empty,_,Empty,_,FlagAss).
	 
grisu_pp_fs_res(FS,Residue,FlagAss) :-
   \+ \+ (deref(FS,Ref,SVs),
          grisu_pp_fs_res_act(Ref,SVs,Residue,FlagAss)).
grisu_pp_fs_res(Ref0,SVs0,Residue,FlagAss) :-
   \+ \+ (deref(Ref0,SVs0,Ref1,SVs1),
          grisu_pp_fs_res_act(Ref1,SVs1,Residue,FlagAss)).

grisu_pp_fs_res_act(Ref,SVs,Residue,FlagAss) :-
  empty_assoc(AssocIn),
  build_iqs(Residue,Iqs,FSResidue),
  (show_res -> residue_args(FSResidue,ResArgs,[Tag-SVs]) ; ResArgs = [Tag-SVs]),
  duplicates_list(ResArgs,AssocIn,DupsMid,AssocIn,VisMid,0,NumMid),
  duplicates_iqs(Iqs,DupsMid,DupsMid2,VisMid,_,NumMid,_),
  % print feature structure in grisu
  grisu_pp_fs(Ref,SVs,DupsMid2,DupsMid3,AssocIn,VisMid2,AssocIn,HDMid,FlagAss),
  % inequalities not yet printed by grisu:
  pp_iqs(Iqs,DupsMid3,DupsOut,VisMid2,VisOut,0,HDMid,HDOut),
  ((show_res,FSResidue\==[])
  -> nl,nl, write('Residue:'), pp_residue(FSResidue,DupsOut,_,VisOut,_,HDOut,_)
  ; true
  ).

grisu_pp_fs(Ref,SVs,Dups0,Dups,Vis0,Vis,HD0,HD,FlagAss) :-
   begin_grisu_output(Stream),
   svs2keyed_feats(SVs,KeyedFeats,Type),
   % We check for a low level hook, i.e. portray_fs here;
   % this hook is e.g. used by the code in save_results.pl
   ( (current_predicate(portray_fs,portray_fs(_,_,_,_,_,_,_,_,_,_)),
     portray_fs(Type,Ref-SVs,KeyedFeats,Vis0,Vis,Dups0,Dups,0,HD0,HD))
   -> true
    ; grisu_portray_fs(Stream,Type,Ref-SVs,KeyedFeats,
                       Vis0,Vis,Dups0,Dups,HD0,HD,FlagAss)
   ).

% ------------------------------------------------------


grisu_portray_fs(Stream,Type,FS,KeyedFeats,
                 Vis0,Vis,Tags0,Tags,HD0,HD,ToplevelFlagAss) :-

   toplevel_init(FS,Tags0,ToplevelFlagAss,HD0,HD1),

   % retrieve the information we stored in the HD arg. between callbacks:
   del_assoc(current,HD1,c(Index0,Index,Mother,Hist0,Hist,FlagAss),HD2),

   deref(FS,Ref,_SVs),
   get_flags(FlagAss,Ref,Flags),

   (get_assoc(Ref,Tags0,Tag),   % This is a structure sharing and
    Mother \= tag(_)            % we were'nt called by writeReentrancyList
   ->  % store the Type/FS/KeyedFeats in Visited under key Ref
       put_assoc(Ref,Vis0,f(Type,FS,KeyedFeats),Vis2),
       addTagToHist(Mother,Tag,Hist0,Hist2),
       writeTag(Stream,Mother,Tag,Index0,Index3,Flags),
       Tags1=Tags0, HD3=HD2
   ;   beginNonTag(Type,Mother,Index0,Index2,Stream,MissingEnds,Flags),
       addToHist(Mother,Type,Index0,Hist0,Hist1),
       put_assoc(Ref,Vis0,not_a_tag,Vis1),
       print_feats(KeyedFeats,Type,Vis1,Vis2,Tags0,Tags1,
                   col(Ref,Index2,Index3,Index0,Hist1,Hist2,
                       FlagAss,Stream),
                   HD2,HD3),
       endNonTag(MissingEnds,Stream) % closes brackets from beginNonTag
   ),
   end_portray(HD0,HD3,HD,Vis2,Vis,Tags1,Tags,Index3,Index,Hist2,Hist,FlagAss,Stream),
   !.

% Catch if something fails in the code above, since else we would
% backtrack into the default pretty printer.
grisu_portray_fs(_,_,_,_,Vis,Vis,Tags,Tags,HD,HD,_) :-
   print_message(warning,'Unexpected failure in pretty printing.'),
   abort.

toplevel_init(FS,Tags0,ToplevelFlagAss,HD0,HD1) :-
   topLevel(HD0), !, 
   % check if a ToplevelFlagAss was provided (i.e. is not a variable):
   (  var(ToplevelFlagAss)
   -> empty_assoc(Empty),
      ToplevelFlagAss = flagAss(_IdentAss,Empty)
   ;  true
   ),
   ToplevelFlagAss = flagAss(IdentAss,UnfAss0),
   get_unfill(FS,Tags0,UnfAss0,UnfAss),
   % The assoc entry "current" we use to memorize between callbacks is:
   % c(Index0,Index,Mother,Hist0,Hist,FlagAss)
   put_assoc(current,HD0,
             c(0,_Index,hd(root),[],_Hist,flagAss(IdentAss,UnfAss)),
             HD1).
   % this put_assoc into HD also ensures we fail topLevel/1 after callback

toplevel_init(_,_,_,HD,HD).

% beginNonTag/7: Ref is not a reentrant node, and mother isn't a tag
% Deal with cases where it's
% a) a list:
beginNonTag(Type,Mother,Index0,Index,Stream,MissingEnds,Flags) :-
   treatList(Type,Mother,Index0,Index,Stream,MissingEnds,Flags), 
   !.

% b) anything else 
% This can have features, so it uses MissingEnds to close paren after
% output of features and their value in the calling predicate.
beginNonTag(Type,_,Index0,Index,Stream,[_],Flags) :-
   beginType(Type,Index0,Index,Stream,Flags).

endNonTag([],_).
endNonTag([_|T],Stream) :-
   end(Stream),
   endNonTag(T,Stream).

writeTag(Stream,Mother,Tag,Index0,Index,Flags) :-
   (   Mother = tl(_)
   ->  begin(Stream,'Z',Index0,Index1,Flags),
       beginTag(Stream,Tag,Index1,Index,Flags),
       end(Stream),             % of beginTag
       end(Stream),             % of begin...'Z'
       end(Stream)              % of mother list
   ;   beginTag(Stream,Tag,Index0,Index,Flags),
       end(Stream)
   ).


end_portray(HD0,HD2,HD,Vis1,Vis,Tags1,Tags,Index3,Index,Hist2,Hist,FlagAss,Stream) :-
   topLevel(HD0),
   !,
   xnl(Stream),xnl(Stream),
   assoc_to_list(Tags1,TagsList),
   writeReentrancyList(TagsList,Vis1,Vis,Tags1,Tags,
                       col(Index3,Index,Hist2,Hist,FlagAss,Stream),
                       HD2,HD3,Stream),
   xnl(Stream),
   writeTree(Stream,Hist,Index,HD3),
   nl(Stream),
   flush_output(Stream),
   HD=HD0.% we exit the top level call with an empty assoc list in HD

% second clause deals with when we're not on topLevel:
end_portray(_,HD,HD,Vis,Vis,Tags,Tags,Index,Index,Hist,Hist,_,_).

writeReentrancyList([],Vis,Vis,Tags,Tags,
                    col(Index,Index,Hist,Hist,_FlagAss,_Stream),
                    HD,HD,_Stream).
writeReentrancyList(Reentrancies0,Vis0,Vis,Tags0,Tags,
                    col(Index0,Index,Hist0,Hist,FlagAss,Stream),
                    HD0,HD,Stream) :-
% This is done inefficiently here and should probably be changed.
% As far as I understand Dale's code, it proceeds as follows:
% 1. select picks out a ref
% 2. get_assoc sees if we've come across this node already and stored 
%    the feature structure under that node
% 3. if get_assoc fails, it backtracks into 1. and selects another
%    ref, until we find one that we know the feature structure of
%    If get_assoc succeeds, we print the value of the reentrancy,
%    by calling portray_fs, wrapping the tag with tag() so that
%    we actually print the value this time (and not just the tag).  
% 4. If step 3 never suceeds, we have a structure sharing that didn't
%    actually occur in the structure. That can occur in case a when 
%    statement, which refers to something our structure was derived 
%    from (and therefore does not have to be reentrant with something 
%    in our structure itself).
   ( select(Ref-Tag,Reentrancies0,Reentrancies),
     get_assoc(Ref,Vis0,f(Type,FS,KeyedFeats)) 
   ->  beginReentr(Stream,Tag,Index0,Index1),
       put_assoc(current,HD0,c(Index1,Index2,tag(Tag),Hist0,Hist1,FlagAss),HD1),
       grisu_portray_fs(Stream,Type,FS,KeyedFeats,Vis0,Vis1,Tags0,Tags1,HD1,HD2,_),
       end(Stream),             % of beginReentr
       xnl(Stream)
   ;   Vis0=Vis1,Tags0=Tags1,Index0=Index2,Hist0=Hist1,HD0=HD2,
       print_message(warning,'Found reentrancy not rooted in this structure.')
   ),
   writeReentrancyList(Reentrancies,Vis1,Vis,Tags1,Tags,
                       col(Index2,Index,Hist1,Hist,FlagAss,Stream),
                       HD2,HD,Stream).


print_feats(Feats,Type,VisIn,VisOut,TagsIn,TagsOut,Col,HDIn,HDOut) :-
   order_features(Feats,OrdFeats),
   print_ord_feats(OrdFeats,Type,VisIn,VisOut,TagsIn,TagsOut,Col,HDIn,HDOut).

print_ord_feats([],_Type,Vis,Vis,Tags,Tags,
            col(_Ref,Index,Index,_Mother,Hist,Hist,_FlagAss,_Stream),
            HD,HD).
print_ord_feats([Feat|Feats],Type,Vis0,Vis,Tags0,Tags,
            col(MotherRef,Index0,Index,Mother,Hist0,Hist,FlagAss,Stream),
            HD0,HD) :-
   print_feat(Feat,Type,Vis0,Vis1,Tags0,Tags1,
              col(MotherRef,Index0,Index2,Mother,Hist0,Hist1,FlagAss,Stream),
              HD0,HD1),
   print_ord_feats(Feats,Type,Vis1,Vis,Tags1,Tags,
               col(MotherRef,Index2,Index,Mother,Hist1,Hist,FlagAss,Stream),
               HD1,HD).


print_feat(fval(F,Val,Restr),Type,Vis0,Vis,Tags0,Tags1,
           col(_,Index0,Index2,Mother,Hist0,Hist1,FlagAss,Stream),
           HD0,HD) :-
   deref(Val,Ref,SVs),            % get the ref of the feature value
   get_flags(F,FlagAss,Ref,Flags),
   beginFeature(Type,F,Index0,Index1,NeedsEnd,Flags,Stream),
% store the type of things with a phon attribute (for tree, in partic. leafs)
  deal_with_phon(F,Restr,Val,Mother,Type,HD0,HD1),
  getNextMother(F,Restr,Type,Mother,Mother1,HD1,HD2),
  put_assoc(current,HD2,c(Index1,Index2,Mother1,Hist0,Hist1,FlagAss),HD3),
% The main recursion into the feature values:
   svs2keyed_feats(SVs,KeyedFeats,XType),
   grisu_portray_fs(Stream,XType,Ref-SVs,KeyedFeats,
                       Vis0,Vis,Tags0,Tags1,HD3,HD,_),
  (   NeedsEnd
  ->  end(Stream)
  ;   true 
  ).

get_flags(Feat,flagAss(IdentAss,UnfAss),Ref,flags(Differ,Unfill)) :-
   get_flags(flagAss(IdentAss,UnfAss),Ref,flags(Differ,TmpUnfill)),
   (   no_write_feat_flag(Feat)	% when user tells us so
   ->  Unfill=true		% unfill a particular feature
   ;   Unfill=TmpUnfill).	% else deal according to its informativeness

get_flags(flagAss(IdentAss,UnfAss),Ref,flags(Differ,Unfill)) :-
   (   (var(IdentAss);		% if variable -> wasn't specified
        get_assoc(Ref,IdentAss,_) )
   ->  Differ=false             % uninterresting = common in diff
   ;   Differ=true 
   ),
   (  get_assoc(Ref,UnfAss,_)       
   -> Unfill=true
   ;  Unfill=false
   ).

deal_with_phon(Feat,Restr,FS,Mother,Type,HD0,HD1) :-
   phon_feat_(Feat),		% user defined as storing the phonology
   sub_type(list,Restr),	% and has appropriate value list
   !,
   deref(FS,TagMid,SVsMid),
   (  fs_string2prolog_list(TagMid-SVsMid,Phon) 
   -> true
   ;  Phon = [none]
   ),
   put_assoc(phon_info(Mother),HD0,Type:Phon,HD1).
deal_with_phon(_,_,_,_,_,HD,HD).

fs_string2prolog_list(Tag-SVs,List) :-
   functor(SVs,Type,_),
   (Type == e_list
   -> List = []
   ;  (Type == list
      ->  List = _
      ;   List = [H|T],
          featval(hd,SVs,Tag,Hd),
          deref(Hd,_TagHd,SVsHd),
	  % remove an atom wrapper if one exists:
	  (  SVsHd = (a_ H)
	  -> true
	  ;  SVsHd = H
	  ),
          featval(tl,SVs,Tag,Tl),
          deref(Tl,TagTl,SVsTl),
          fs_string2prolog_list(TagTl-SVsTl,T))).


treatList(ne_list,Mother,Index0,Index,Stream,[],Flags) :- 
   !,
   (   Mother = tl(_)
   ->  Index is Index0 + 1      % We want a different id
                                % than for tl:hd for use in
                                % the assoc lists. 
   ;   begin(Stream,'L',Index0,Index,Flags)  
       % the end for this begin is in next two clauses
   ).
treatList(e_list,Mother,Index0,Index,Stream,[],Flags) :- 
   !,
   (   Mother = tl(_)
   ->  end(Stream),             % of mother list 
       Index=Index0
   ;   begin(Stream,'L',Index0,Index,Flags),
       end(Stream)
   ).

% The hd and the tl of a list are displayed by the special
% [...] notation, so we never unfill those attribute values.
% The argument of flags/2 in the following two clause thus
% is always set to false.

% Type = list or other value that can occur as value of tl attribute
treatList(Type,tl(_),Index0,Index,Stream,[_,_,_],flags(Differ,_)) :-
   begin(Stream,'Z',Index0,Index1,flags(Differ,false)),
   beginType(Type,Index1,Index,Stream,flags(Differ,false)).
   % end of beginType -> one true in needsEnd continuation
   % end of begin ... Z -> one true in needsEnd continuation
   % end of mother list -> one true in needsEnd continuation

treatList(Type,hd(_),Index0,Index,Stream,[_],flags(Differ,_)) :-
   % Note that we never unfill the tl of a list since lists 
   % are displayed through the special notation, so the second 
   % argument of flags/2 is always specified to be false.
   beginType(Type,Index0,Index,Stream,flags(Differ,false)).
   % end of beginType -> one true in needsEnd continuation
   

% When the pretty printing hook is called for the first time in
% printing a structure, the HDIn Argument is an empty_assoc (i.e. t).
topLevel(A) :-
   empty_assoc(A).

addTagToHist(none,_,Hist,Hist) :- !.
addTagToHist(Mother,Tag,Hist,[Mother=tag(Tag)|Hist]) :- !.   

addToHist(tag(N),_Type,Index,Hist,[tag(N)-Index|Hist]).
addToHist(hd(I),_Type,Index,Hist,[I-hd(Index)|Hist]).
addToHist(tl(I),Type,Index,Hist,[Tail|Hist]) :-
   makeTail(Type,I,Index,Tail),
   !.
addToHist(tl(_I),_Type,_Index,Hist,Hist).
addToHist(dtrs(I),Type,Index,Hist,[Dtrs|Hist]) :-
   makeDtrs(Type,I,Index,Dtrs).
addToHist(none,_Type,_Index,Hist,Hist).

makeTail(ne_list,I,Index,I-tl(Index)).
makeTail(list,I,_Index,I-tl(list)).
makeTail(e_list,I,_Index,I-tl(e_list)).

makeDtrs(ne_list,I,Index,I-dtrs(Index)).
makeDtrs(list,I,_Index,I-dtrs(list)).
makeDtrs(e_list,I,_Index,I-dtrs(e_list)).

getNextMother(hd,_,ne_list,Index,hd(Index),HD,HD) :- !.
getNextMother(tl,_,ne_list,Index,tl(Index),HD,HD) :- !.
getNextMother(Feat,Restr,Type,Index,dtrs(Index),HD0,HD1) :- 
   dtrs_feat_(Feat),		% user defined as storing the dtrs
   sub_type(list,Restr),	% and has appropriate value list
   !,
   put_assoc(tree_info(Index),HD0,Type,HD1). % store tree mother info by Index
getNextMother(_,_,_,_,none,HD,HD).

% Note on getNextMother: We explicitly test for the type ne_list since 
% attributes hd and/or tl could be appropriate for other types as well,
% but we only have a special list syntax for type ne_list.

get_output_stream(Stream) :-
  (   \+ grisu_debug_flag,
      current_predicate(socket_info,socket_info(_,_)),
      socket_info(_,Stream)
  ->  true
  ;   Stream = user_output
  ).


% beginType(+Type:<typename>,
%           +CurrentIndx:<integer>,-NextIndex:<integer>,
%           +Stream:<stream>,+Flags)

% a) a_ atoms
beginType((a_ Atom),Index0,Index,Stream,Flags) :-
   !,
   begin(Stream,'A',Index0,Index,Flags), 
   writeqt(Stream,Atom).

% b) ordinary types
beginType(Type,Index0,Index,Stream,Flags) :-
   begin(Stream,'S',Index0,Index1,Flags),
   write(Stream,'('),
   writeFlags(Stream,Flags),
   writeIndex(Stream,Index1,Index),
   writeqt(Stream,Type),
   end(Stream).

% Note on beginFeature: We explicitly test for the type ne_list since 
% attributes hd and/or tl could be appropriate for other types as well,
% but we only have a special list syntax for type ne_list.

beginFeature(ne_list,_Feat,Index, Index,false,_Flags,_Stream) :- !.
beginFeature(_Type,   Feat,Index0,Index,true,  Flags, Stream) :-
   begin(Stream,'V',Index0,Index,Flags),
   writeFeat(Stream,Feat).

% begin a tag
beginTag(Stream,Num,Index0,Index,Flags) :-      % new 1.0.6 syntax
   begin(Stream,'#',Index0,Index,Flags),
   write(Stream,' '),           % new syntax: space required
   write(Stream,Num).

% begin a reentr
beginReentr(Stream,Num,Index0,Index) :- % new 1.0.6 syntax
   begin(Stream,'R',Index0,Index),
   write(Stream,' '),           % new syntax: space required
   write(Stream,Num).


% begin general cases: X is one of the different node kinds
begin(Stream,X,Index0,Index) :-
   begin(Stream,X,Index0,Index,false).
begin(Stream,X,Index0,Index,Flags) :-
   write(Stream,'('),
   write(Stream,X),
   writeFlags(Stream,Flags),
   writeIndex(Stream,Index0,Index).

end(Stream) :- 
   write(Stream,')').

writeFlags(Stream,Flags) :-
   (   (nonvar(Flags),Flags=flags(Differ,Unfill))
   ->  (Differ -> write(Stream,'*');true),
       (Unfill -> write(Stream,'+');true) 
   ; true).

writeIndex(Stream,Index0,Index) :-
   write(Stream,Index0),        % new syntax: index no longer quoted
   Index is Index0+1.

writeqt(Stream,X) :-
   write(Stream,'"'),
   write(Stream,X),
   write(Stream,'"').


writeFeat(Stream,Feat):-
   writeqt(Stream,Feat). % we send the feat as is, i.e. no capitalize_atom 

writeTree(Stream,Hist,Index0,HD) :-
   removeTagsFromHistory(Hist,HistNoTags),
   vertices_edges_to_ugraph([],HistNoTags,Graph),
   bindListsInHist(Graph,T0),
   ( select(dtrs(0,_),T0,_) 
   -> writeTree(Stream,0,T0,_T,Index0,_Index,HD)
   ;  true % no top-level tree found, so none is displayed
   ).

% ----------------------------------------------------------
% removeTagsFromHistory/2: connects all paths to their
% actual values, i.e. in case the value is a tag inserts the
% value that tag is pointing to.
% Example:
% | ?- removeTagsFromHistory([tag(0)-11,5-tl(e_list),hd(5)=tag(0),3-tl(5),hd(3)=tag(0)],L).
%      L = [5-tl(e_list),5-hd(11),3-tl(5),3-hd(11)] ? 

removeTagsFromHistory(Hist0,Hist) :-
   empty_assoc(Bind0),
   store_bindings(Hist0,Hist1,Bind0,Bind1),
   apply_bindings(Hist1,Hist,Bind1).

store_bindings([],[],Bind,Bind).
store_bindings([tag(N)-S|R0],R,Bind0,Bind) :-
   !,
   put_assoc(N,Bind0,S,Bind1),
   store_bindings(R0,R,Bind1,Bind).
store_bindings([F|R0],[F|R],Bind0,Bind) :-
   store_bindings(R0,R,Bind0,Bind).

apply_bindings([],[],_Bind).
apply_bindings([Term=tag(N)|Hist0],[Index-NewTerm|Hist],Bind) :-
   !,
   get_assoc(N,Bind,S), 
   Term =.. [What,Index],
   NewTerm =.. [What,S],
   apply_bindings(Hist0,Hist,Bind).
apply_bindings([F|Hist0],[F|Hist],Bind) :-
   apply_bindings(Hist0,Hist,Bind).

% ----------------------------------------------------------

bindListsInHist(Hist0,Hist) :-
   empty_assoc(Binding),
   put_assoc(e_list,Binding,[],Binding1),
   put_assoc(list,Binding1,_,Binding2),
   bindListsInHist(Hist0,Hist,Binding2).

bindListsInHist([],[],_Binding).
bindListsInHist([N-[hd(H),tl(TL0)]|Hist0],Hist,Binding0) :- !,
   makeNewTail(TL0,TL,Binding0,Binding1),
   get_binding(N,Var,Binding1,Binding2),
   Var=[H|TL],
   bindListsInHist(Hist0,Hist,Binding2).
bindListsInHist([N-[dtrs(D)]|Hist0],[dtrs(N,Dtrs)|Hist],Binding0) :- !,
   get_binding(D,Dtrs,Binding0,Binding1),
   bindListsInHist(Hist0,Hist,Binding1).   
bindListsInHist([_F|Hist0],Hist,Binding0) :- !,
   bindListsInHist(Hist0,Hist,Binding0).

get_binding(S,BoundTo,Binding0,Binding) :-
   get_assoc(S,Binding0,BoundTo), !,
   Binding = Binding0.
get_binding(S,BoundTo,Binding0,Binding) :-
   put_assoc(S,Binding0,BoundTo,Binding).

makeNewTail(e_list,[],Binding,Binding) :- !.
makeNewTail(list,_,Binding,Binding) :- !.
makeNewTail(N,TL,Binding0,Binding) :-
   get_binding(N,TL,Binding0,Binding).

% ----------------------------------------------------------

writeTree(Stream,N,T0,T,Index0,Index,HD) :-
   begin(Stream,'T',Index0,Index1),
   ( select(dtrs(N,Dtrs),T0,T1)
   ->  true
   ;   Dtrs=[], T1=T0
   ),
   % node label:
   writeNodeLabel(Stream,Dtrs,N,HD),
   % arc label:
%   writeqt(Stream,some-feat),
   % link id (no longer quoted):
   write(Stream,N),
   writeDtrs(Dtrs,T1,T,Index1,Index,Stream,HD),
   end(Stream).

writeNodeLabel(Stream,_Dtrs,N,HD) :-
   (get_assoc(phon_info(N),HD,Info) 
   -> writeqt(Stream,Info)
   ;  (get_assoc(tree_info(N),HD,Info)
      -> writeqt(Stream,Info)
      ;  writeqt(Stream,'no label'))).

writeDtrs([],T,T,Index,Index,_Stream,_HD).
writeDtrs([N|Rest],T0,T,Index0,Index,Stream,HD) :-
   writeTree(Stream,N,T0,T1,Index0,Index1,HD),
   writeDtrs(Rest,T1,T,Index1,Index,Stream,HD).

xwrite(Stream,Term0) :-
   rm_nl(Term0,Term),
   write(Stream,Term),
   write(Term).


% For xnl/1 and xtab/1 below, I commented out to get screen output
% which can directly be loaded into grisu. Comment back in to human
% readable output on screen. (wdm, 1.August.02)

%xnl(user_output) :- !,
%   nl(user_output).
xnl(_).

%xtab(user_output,N) :- !,
%   tab(user_output,N).
xtab(_,_).


% ========================================================


begin_grisu_output(Stream) :-
   get_output_stream(Stream),
   format(Stream,'!newdata',[]),
   (  pref_struc_flag 
   -> format(Stream,'(structure)',[])
   ;  true
   ),
   get_title(Title),
   format(Stream,'"~w"',[Title]).

% ========================================================
% specific portray hooks
% ========================================================

portray_lex(Word,Tag,SVs,Residue) :-
   grisu_flag,			% fails immediately if grisu not on
   set_title(lex:Word),
   grisu_pp_fs_res(Tag,SVs,Residue,_),
   clear_title.

% ------------------------------------------------------
% portray_cat/5: basic feature structure output
% called by rec/1, rec/2, rec_best/2, rec_list/2, gen/1, gen/2 

portray_cat(Words,Desc,Tag,SVs,Residue) :-
   grisu_flag,			% fails immediately if grisu not on
   (  Desc = bot 
   -> set_title(Words)
   ;  set_title(Words-Desc)
   ),
   grisu_pp_fs_res(Tag,SVs,Residue,_),
   clear_title.

% ------------------------------------------------------

portray_mgsat(Desc,Ref,SVs,Residue) :-
   grisu_flag,			% fails immediately if grisu not on
   set_title(mgsat:Desc),
   grisu_pp_fs_res(Ref,SVs,Residue,_),
   clear_title.

% ------------------------------------------------------

portray_ale_macro(MacroName,_Desc,Tag,SVs,Residue) :-
   grisu_flag,
   set_title(macro:MacroName),
   grisu_pp_fs_res(Tag,SVs,Residue,_),
   clear_title.

% ------------------------------------------------------

portray_empty(I,Tag,SVs,_Dtrs,RuleName,ND,Residue) :-
   grisu_flag,
   set_title('Empty category'),
   grisu_pp_fs_res(Tag,SVs,Residue,_),
   (no_interpreter 
    -> true
     ; nl, write('     index: '),(functor(I,empty,2)
                                 -> arg(1,I,E),
                                    write(E)
                                  ; write(I)),
       nl, write('      rule: '),write(RuleName),
       nl, write(' # of dtrs: '),write(ND),
       nl
    ),
    clear_title.

% ------------------------------------------------------

% portray_unif_failure(Ref1,SVs1,Frozen1,Ref2,SVs2,Frozen2) :-
% portray_path_failure(Path1,Path2,Tag,SVs,Frozen) :-
% portray_feat_failure(Feat,Ref,SVs,Frozen) :-
% portray_macro_failure(MacroName,Ref,SVs,Frozen) :-
% portray_addtype_failure(Type,Ref,SVs,Frozen) :-
% portray_undef_type(Atom,Ref,SVs,Frozen) :-
% portray_desc_failure(Desc,Ref,SVs,Frozen) :-
% portray_featpath_failure(Feat,Path,Tag,SVs,Frozen) :-
% portray_edge_discard(LReln,I,Left,Right,Tag,SVs,RuleName,ND,Res) :-
% portray_edge_retract(I,Left,Right,ETag,ESVs,ERuleName,NED,ERes) :-
% portray_incoming_edge(Left,Right,Tag,SVs,RuleName,ND,Res) :-
% portray_edge(pending,Left,Right,TagOut,SVsOut,RuleName,ND,Res) :-
% portray_dtr_edge(D,DLeft,DRight,DTag,DSVs,DRule,DND,DRes) :-
% portray_ale_goal(Goal,Residue) :-
% portray_edge(I,M,N,Tag,SVs,RuleName,ND,Residue) :-
% portray_lex_rule(RuleName,Desc1,Desc2,Tag1,bot,Tag2,bot,Residue,Goal,Morphs) :-
% portray_ale_clause(Head,Body) :-
% portray_rule(TagMoth,bot,Dtrs,Residue) :-

% ------------------------------------------------------

portray_approps(Type,FVs) :-
   grisu_flag,                  % fails immediately if flag not on
   unfill_off,
   get_output_stream(Stream),
   format(Stream,'!newdata"show_approp(~w)"',[Type]),
   Index0=1,
   beginType(Type,Index0,Index1,Stream,false),
   pa_feature_values(FVs,Index1,_FinalIndex,Stream),
   end(Stream),                 %'S'
   nl(Stream),
   flush_output(Stream),
   unfill_restore.

pa_feature_values([],Index,Index,_Stream).
pa_feature_values([Feat:Type|Rest],Index0,Index,Stream) :-
   pa_feature_value(Feat,Type,Index0,Index1,Stream),
   pa_feature_values(Rest,Index1,Index,Stream).

pa_feature_value(Feat,Type,Index0,Index,Stream):-
   begin(Stream,'V',Index0,Index1),
   writeFeat(Stream,Feat),
   beginType(Type,Index1,Index,Stream,false),
   end(Stream),                 % Type   
   end(Stream).                 %'V'

% ========================================================
% redefinition of pp_residue/7 until output of residues
% is integrated in ale.pl in a way allowin output in grisu

% Compared to the original in ale.pl, the only change in this predicate 
% is to switch grisu off for residue output so that the 
% default pretty printer handles them

:- abolish(pp_residue/7).

pp_residue(Residue,Dups0,Dups,Vis0,Vis,HD0,HD) :-
   grisu_off,
   original_pp_residue(Residue,Dups0,Dups,Vis0,Vis,HD0,HD),
   grisu_on.

% the original definition from ale.pl:
original_pp_residue([],Dups,Dups,Vis,Vis,HD,HD).
original_pp_residue([_-Goal|Residue],DupsIn,DupsOut,VisIn,VisOut,HDIn,HDOut) :-
  pp_resgoal(Goal,DupsIn,DupsMid,VisIn,VisMid,HDIn,HDMid),
  original_pp_residue(Residue,DupsMid,DupsOut,VisMid,VisOut,HDMid,HDOut).


% ========================================================
%% -*-Mode: prolog;-*-

