%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%   $RCSfile: graphviz.pl,v $
%%  $Revision: 1.13 $
%%      $Date: 2003/10/02 22:24:49 $
%%     Author: Detmar Meurers (dm@ling.osu.edu)
%%    Purpose: Provides graphical hierarchy output
%%             using graphviz package, cf.
%%             http://www.research.att.com/sw/tools/graphviz/
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Log: graphviz.pl,v $
%% Revision 1.13  2003/10/02 22:24:49  dm
%% Catch and output special type a_ _ as a string to keep the two symbols
%% together in output. Also changed to use system/1 instead of shell/1
%% calls now to use default shell.
%%
%% Revision 1.12  2003/10/02 21:14:15  dm
%% Added grisu as being usable with graphviz_option(svg,grisu) - from
%% Version 1.4.0 on it can display svg files!
%%
%% Revision 1.11  2003/09/29 14:48:18  dm
%% Correction of last check in: I wanted gv not ghostview as default for
%% viewing.
%%
%% Revision 1.10  2003/09/29 01:28:57  dm
%% Changed graphviz_option_default to ps and ghostview since ps includes
%% text as text instead of as graphics and ghostview supports scaling and
%% panning in lager-than-screen images well.
%%
%% Revision 1.9  2003/05/13 04:58:49  dm
%% Allow possibility of intro/2 being part of a subs/2 declaration (cf. the
%% old ALE style of declaring a signature)
%%
%% Revision 1.8  2003/05/13 04:28:27  dm
%% Added faile_if_undefined/1 around intro/2 to catch case in which grammar
%% doesn't contain any.
%%
%% Revision 1.7  2003/05/13 04:26:30  dm
%% simplified draw_local/4
%%
%% Revision 1.6  2003/05/13 04:13:08  dm
%% Added draw_hierarchy/0 and draw_signature/0 to output entire
%% hierarchy.
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% provides the following user commands:

draw_hierarchy(Type) :-
   draw_hierarchy(Type,false).

draw_signature(Type) :-
   draw_hierarchy(Type,true).

% and two special versions to output the entire hierarchy:

draw_hierarchy :-
   draw_hierarchy(bot).

draw_signature :-
   draw_signature(bot).

% ------------------------------------------------------
% The predicate graphviz_option/2 can be defined by
% the user as part of the grammar to specify which format 
% is supposed to be generated and which program can view
% that format:

graphviz_option_(Format,Program) :-
   (   current_predicate(graphviz_option,graphviz_option(_,_)),
       graphviz_option(UserFormat,UserProgram)
   ->  ((atom(UserFormat),atom(UserProgram)) 
       ->  UserFormat=Format,UserProgram=Program
       ;   graphviz_option_default(Format,Program),
	   format(user_error,'~N**Error: Both arguments of graphviz_option(<format>,<program>) must be atoms!~n(surround ~w and ~w by single quotes to obtain atoms)~nUsing fallback graphviz_option(~w,~w).~n',[UserFormat,UserProgram,Format,Program])
       )
   ;  graphviz_option_default(Format,Program)
   ).

%% The default if the user hasn't defined graphviz_option/2
%graphviz_option_default(jpg,xv).
graphviz_option_default(ps,gv).

% -----------------------------------------------------------------------

:- use_module(library(system),[tmpnam/1,system/1]).
:- use_module(library(charsio),[format_to_chars/3,format_to_chars/4]).

draw_hierarchy(Type,ShowApprops) :-
	type_hierarchy_compiler:ensure_signature_present,
	type_hierarchy_compiler:ensure_type_defined(Type),
	type_hierarchy_compiler:get_all_subtypes(Type,Result,[],_),
	(Result == [] ->
	    format(user_output,'**Error: Type "~w" has no subtypes.~n',[Type])
	;
	    tmpnam(Filename),
	    format_to_chars('~w.dot',[Filename],DotFilenameS),
	    atom_codes(DotFilename,DotFilenameS),

	    tell(DotFilename),
	    format('digraph ~w {~n',[Type]),

            % horizontal space between nodes in inches:
	    format('ranksep="0.4 equally";~n',[]), 

            % horizontal space between nodes in inches:
	    format('nodesep="0.2 equally";~n',[]), 

	    % justification of labels (relevant if it has more lines)
	    format('labeljust="l";~n',[]),

            % set fontpath
%	    format('fontpath="/usr/share/fonts/default/TrueType/";~n',[]),

	    % font name and size to use
%	    format('node [fontname="helr____"];~n',[]),

	    % nodes are just text and there's no padding around them
	    format('node [height=0,width=0,shape=plaintext];~n',[]),

	    % here we can compress it to a certain size
%	    format('ratio="compress";~nsize="8.5,11";~n',[]),

	    % or one can just specify a certain size, without changing the ratio
%	    format('size="13,10";~n',[]),

            % or one can set size in inches at which to cut into pages, and margins
%	    format('page="8.5,11"~numargin="0.1,0.1"~n',[]), 

	    % no arrows at end of edges:
	    format('edge [arrowhead="none"];~n',[]), 

	    draw_hierarchy([Type-Result],_,ShowApprops),

	    format('~N}~n',[]),
	    told,

	    graphviz_option_(Format,Program),

	    (  Program = grisu
	    -> (  grisu_flag
	       -> format_to_chars('dot -Tsvg ~w > ~w.svg',[DotFilename,Filename],String),
		  atom_codes(Command,String),
%	          format('~N~p~n',[Command]),	% tell user what we're doing
	          system(Command),
		  get_output_stream(Stream),
		  format(Stream,'~N!show_svg_file "~w.svg"~n',[Filename]),
		  flush_output(Stream)	       
	       ;  format(user_error,'~N**Error: No grisu Interface present. Specify other graphviz_option/2 or run grisu.~n',[]),
		  fail
	       )
	    ;  format_to_chars('dot -T~w ~w > ~w.~w;~w ~w.~w&',[Format,DotFilename,Filename,Format,Program,Filename,Format],String),
	       atom_codes(Command,String),
	       format('~N~p~n',[Command]),	% tell user what we're doing
	       system(Command)
	    )).

% print subtypes in style of signature input format
draw_hierarchy([],_,_).
draw_hierarchy([H|T],Mother,ShowApprops) :- 
   (   H=(Type-Subtypes)
   ->  draw_local(Type,Type1,Mother,ShowApprops),
       draw_hierarchy(Subtypes,Type1,ShowApprops)
   ;   draw_local(H,_,Mother,ShowApprops)
   ),
   draw_hierarchy(T,Mother,ShowApprops).

draw_local(Type,Type1,Mother,ShowApprops) :-
   (  Type=multi(Type1)		% eliminate multi wrappers
   -> true
   ;  Type=Type1),
   (  Type1=(a_ _)		% show a_ atoms as anon var string
   -> Type2= '"a_ _"'
   ;  Type1=Type2),
   (   var(Mother)		           % if top level
   ->  true			           % do nothing
   ;   format('~w -> ~w;~n',[Mother,Type2]) % else output local graph
   ),
   (  (ShowApprops, 
	  (fail_if_undefined(intro(Type2,FVs))
	  ;fail_if_undefined(subs(Type2,intro(_SubT,FVs)))) )
   ->  format('~w [label="~w\l',[Type2,Type2]),
       draw_approps(FVs),
       format('"];~n',[])
   ;   true
   ).

draw_approps([]).
draw_approps([H|T]) :-
   format('   ~w\l',[H]),
   draw_approps(T).
