%% $Id: arc2vcg.pl,v 1.6 2002/12/07 18:26:24 dm Exp $

:- use_module(library(system),[system/1]).

:- dynamic get_arc/4.

% pruned_global_automaton_filename/1 contains the
% name of the file containing the global automaton
% (defined in main.pl)

%Top_level_calls: lr_show_global/0 and lr_show_automata/0

lr_show_global :- 
   retractall(get_arc(_,_,_)),
   assert((get_arc(From,To,Label) :- arc(From,To,Label))),
   pruned_global_automaton_filename(Filename),
   lr_show_aux(Filename).

lr_show_automata :-
   lr_show_automaton(_),
   fail.
lr_show_automata.

lr_show_automaton(Phon) :-
   lex_info(Phon,Id,_Lex),
%%%   fail_if_undefined((Phon ---> Pred)),
%%%   functor(Pred,FsaID,_),
%%%   name(FsaID,[_Q|T]),		% drop the q to get the number
%%%   name(Id,T),
   retractall(get_arc(_,_,_)),
   assert((get_arc(From,To,Label) :- lex_arc(Id,From,To,Label))), 
   construct_filename(Phon,Id,Filename),
   lr_show_aux(Filename).

lr_show_aux(Filename) :-
   begin_vcg_file(Filename,Stream,VcgFilename),
   write_vcg_automaton(Stream),
   end_vcg_file(Stream),
   call_vcg(VcgFilename).

% The input for lr_show_global consists of asserted arcs of the form:
% arc(from-id, to-id, label, cycle-flag)
% Every automaton starts at node 0.

% The input for lr_show_automaton/1 consists of asserted arcs of the form:
% lex_arc(id-of-entry,from-id, to-id, label, cycle-flag)
% Every automaton starts at node 0.


begin_vcg_file(Filename,Stream,FullFilename) :-
    my_concat(Filename,'.vcg',FullFilename),
    open(FullFilename,write,Stream),
    write(Stream,'graph: { title:"test"'),nl(Stream),
    write(Stream,'xbase: 50'),nl(Stream),
    write(Stream,'ybase: 50'),nl(Stream),
    write(Stream,'yspace: 40'),nl(Stream),
    write(Stream,'splines: yes'),nl(Stream),
    write(Stream,'port_sharing: no'),nl(Stream),
    write(Stream,'layoutalgorithm: maxdepth'),nl(Stream),
    write(Stream,'orientation: left_to_right'),nl(Stream),
    write(Stream,'arrow_mode: free'),nl(Stream),
    write(Stream,'near_edges: yes'),nl(Stream),
    write(Stream,'display_edge_labels: yes'),nl(Stream).

end_vcg_file(Stream) :-
   write(Stream,'}'),
   close(Stream).

call_vcg(Filename) :-
    name('xvcg -silent -nocolors ',L1),
    name(Filename,L2),
    append(L1,L2,L3),
    name(' &',L4),
    append(L3,L4,L),
    name(Call,L),
    system(Call).
     
% The input consists of asserted arcs are of the form:
% arc(from-id, to-id, label, cycle-flag)
% Every automaton starts at node 0.

write_vcg_automaton(Stream) :-
   write_vcg_edges(Stream,Nodes),
   write_vcg_startnode(Stream),
   sort(Nodes,SortedNodes),	% sort, also removes duplicates
   write_vcg_nodes(SortedNodes,Stream).

write_vcg_edges(Stream,Nodes) :-
   findall(To,do_one_arc(Stream,_From,To),Nodes).

do_one_arc(Stream,From,To) :-
   get_arc(From,To,Label),
   write_one_vcg_edge(Stream,From,To,Label).

write_one_vcg_edge(Stream,From,To,Label):-
    write(Stream,'edge: { sourcename: "'),
    write(Stream,'q'),
    write(Stream,From),
    write(Stream,'" targetname: "'),
    write(Stream,'q'),
    write(Stream,To),
    write(Stream,'" label: "'),
    write(Stream,Label),
    write(Stream,'"}'),
    nl(Stream).


write_vcg_nodes([],_Stream).
write_vcg_nodes([H|T],Stream) :-
   write_vcg_node(H,Stream),
   write_vcg_nodes(T,Stream).


write_vcg_startnode(Stream) :-
   write(Stream,'node: { title: "'),
   write(Stream,'q'),
   write(Stream,0),
   write(Stream,'" label: "'),
   write(Stream,'q'),
   write(Stream,0),
   write(Stream,'" shape: triangle}'),
   nl(Stream).    

write_vcg_node(Name,Stream):-
   write(Stream,'node: { title: "'),
   write(Stream,'q'),
   write(Stream,Name),
   write(Stream,'" label: "'),
   write(Stream,'q'),
   write(Stream,Name),
   write(Stream,'" shape: ellipse}'),
   nl(Stream).  

construct_filename(Phon,Id,Filename) :-
   my_concat('automaton-id',Id,Name1),
   my_concat('-',Name1,Name2),
   my_concat(Phon,Name2,Filename).
