% $Id: convert_arcs.pl,v 1.3 2002/08/08 19:11:59 dm Exp $

:- dynamic count/1.
:- dynamic state_arc/5.
:- dynamic arc/4.

% convert-arcs/0
% makes the following changes to the database

% input:  arc(ArcID,RuleID,Prefix,Follow,Cycle)

% intermediate:  state_arc(StateInID,StateOutID,RuleID,Follow,Cycle)

% output:  arc(StateInID,StateOutID,RuleID,Cycle)

convert_arcs :-
    retractall(arc(_,_,_,_)),
    retractall(state_arc(_,_,_,_,_)),
    retractall(count(_)),
    (   follow_arc(_ArcID,RuleID,[],Follow,no),
	StIn = 0,
	(   retract(count(N)) ->
	    StOut is N + 1
	;   StOut is 1
	),
	assert(count(StOut)),
	assert(state_arc(StIn,StOut,RuleID,Follow,no)),
%	portray_clause(state_arc(StIn,StOut,RuleID,Follow,no)),
	next_states(Follow,StOut),
	(   retract(state_arc(CycIn,Rule2ID,Rule2ID,Follow,yes)) ->
	    assert(state_arc(CycIn,StOut,Rule2ID,Follow,yes))
%	    portray_clause(state_arc(CycIn,StOut,Rule2ID,Follow,yes))
	;   true
	),
	fail
    ;   true
    ),
    pruned_global_automaton_filename(Filename),
    my_concat(Filename,'.pl',PlFilename),
    open(PlFilename,write,Stream),
    elim_follows(Stream),
    close(Stream).

% next_states(+Follow,+StIn)
% takes the (ArcID,RuleID) pairs in Follow and the previous state (StIn)
% to construct new state_arcs without prefixes and with State IDs

next_states([],_StIn).
next_states([(ArcID,RuleID)|Rest],StIn) :-
    next_states(Rest,StIn),
    (   follow_arc(ArcID,RuleID,_Prefix,Follow,Cycle) ->

	% if there is a cycle, we add a pointer in the form of an
	% identical RuleID instead of the StateOut
	% this RuleID will necessarily be identical with the state
	% that the cycle is returning to -- and it will be the first
	% state that we reach going backwards

	(   Cycle = yes ->
%	    portray_clause(state_arc(StIn,RuleID,RuleID,Follow,Cycle)),
	    assert(state_arc(StIn,RuleID,RuleID,Follow,Cycle))
        ;   retract(count(N)),
	    StOut is N + 1,
	    assert(count(StOut)),
	    next_states(Follow,StOut),
	    assert(state_arc(StIn,StOut,RuleID,Follow,Cycle)),
%	    portray_clause(state_arc(StIn,StOut,RuleID,Follow,Cycle)),
	    (   retract(state_arc(CycIn,Rule2ID,Rule2ID,Follow,yes)) ->
		assert(state_arc(CycIn,StOut,Rule2ID,Follow,yes))
%	        portray_clause(state_arc(CycIn,StOut,Rule2ID,Follow,yes))
	    ;   true
	    )
	)
	% we need this true condition because the ArcID arc may no longer
	% exist (it was pruned from the automaton, but not removed from
	% the Follow list)  So, we just ignore it and move on
    ;   true
    ).

% elim_follows/0
% takes the state_arcs and removes the Follow lists from them

elim_follows(Stream) :-
    state_arc(StIn,StOut,RuleID,_,Cycle),
    assert(arc(StIn,StOut,RuleID,Cycle)),
    portray_clause(Stream,arc(StIn,StOut,RuleID,Cycle)),
    fail.
elim_follows(_).
