% $Id: main.pl,v 1.18 2003/11/19 22:36:34 dm Exp $
%
% This is the main load file for the lexical rule compiler
% described in Meurers/Minnen (1997).
%
% Authors: Vanessa Metcalf, Detmar Meurers, Markus Dickinson

:- use_module(library(system),[system/1,datime/1]).

:- trale_search_path_add(lr_compiler,trale_home(lr_compiler)).

:- op(1100,xfx,'===').

:- dynamic lr_debug/0, debug_file/1, running_lr_compiler/0.

:- ensure_loaded(lr_compiler(lr_utils)).
:- ensure_loaded(lr_compiler(step_one)).
:- ensure_loaded(lr_compiler(step_two)).
:- ensure_loaded(lr_compiler(step_three)).
:- ensure_loaded(lr_compiler(step_four)).
%%%:- ensure_loaded(lr_compiler(unfold)).
:- ensure_loaded(lr_compiler(arc2vcg)).

% FileIn contains the base lexicon and the lexical rules, as well as
% a signature/1 statement if the signature file isn't called signature

% defining filename to be used for output
lr_compiler_output_filename('lr_compiler_output.pl').

compile_lrs(FileIn) :-
   lr_compiler_output_filename(FileOut),
   assert(running_lr_compiler),
   % loads and compiles signature and relevant part of grammar
   compile_gram_special(FileIn),!,
   retractall(running_lr_compiler),
   format(user_output,'~NCompiling lexical rules...',[]),
   flush_output(user_output),
   (lr_debug 
   -> open('lr_debug_out',write,DebugStream),
      assert(debug_file(DebugStream)) 
   ;  true),
   tell(FileOut),
   datime(datime(Y,M,D,H,Min,S)),
   format('% File generated by compile_lrs(~p,~p)~n% at ~p/~p/~p ~p:~p:~p~n~n',[FileIn,FileOut,Y,M,D,H,Min,S]),
   compile_lrs,
   told,
   format(user_output,'~nLexical rule compiler output in file "~p".~n~n',[FileOut]),
   (lr_debug 
   -> close(DebugStream)
   ;  true).

compile_lrs :-
   format(':- multifile fun/1.~n:- multifile if/2.~n~n',[]),
   format(user_output,'1,',[]),flush_output(user_output),
   step_one,
   format(user_output,'2,',[]),flush_output(user_output),
   step_two,
   format(user_output,'3,',[]),flush_output(user_output),
   step_three,
   format(user_output,'4',[]),flush_output(user_output),
   step_four,
   format(user_output,'.~n',[]).
%%%    unfold.


% compile_gram_special is just like compile_gram but without
% a) compile_grammar_act
% b) compile_cons_act
% since we don't need the compiled phrase structure rules
% and the type constraints would interfere with the code compilation

compile_gram_special(File) :-
  abolish_preds,
  reconsult(File),
  compile_gram_special.

compile_gram_special :-
  touch('.alec_throw'),
  absolute_file_name('.alec_throw',AbsFileName),
  retractall(ale_compiling(_)),
  assert(ale_compiling(AbsFileName)),
  compile_sig_act,
  compile_fun_act,
  % do cleanup normally starting compile_cons_act/0:
  abolish((trale_cons)/2),abolish((cons)/2),abolish(('*>')/2),
  compile_logic_act,
  compile_subsume_act,
  compile_dcs_act,
%  compile_grammar_act,
  retract(ale_compiling(_)).

% --------------------------------------------------------
% Define predicate for user to include in grammar to load input or output
% of lexical rule compiler


lr_compiler_input(Input) :-
	(  running_lr_compiler
        -> (  is_list(Input)
           -> InFiles=Input
           ;  InFiles=[Input]
           ),
	   InFiles 
        ;  lr_compiler_output_filename(OutFile),
           [OutFile]
        ).

% --------------------------------------------------------
% Define the name to be used for the global automaton output:

pruned_global_automaton_filename(global_fsa).

% --------------------------------------------------------
% add flag and support for debugging output to files

show_lr_debug :-
  assert(lr_debug).
hide_lr_debug :-
  retractall(show_res).

%:- show_lr_debug.

% lr_assert asserts and if lr_debug is true also writes it out
lr_assert(X) :-
   assert(X),
   (lr_debug -> debug_file(S),portray_clause(S,X) ; true).

% --------------------------------------------------------
% Banner
% --------------------------------------------------------

:- nl,write('
Lexical Rule Compiler Version 1.0
Copyright (C) 2002, Vanessa Metcalf, Detmar Meurers
All rights reserved'),nl,nl.
