%%% ==================================================================
%%% $Id: socket_interface.pl,v 1.4 2003/11/22 03:20:01 dm Exp $
%%%
%%% Communication over sockets: a prolog client
%%%
%%% Author: Detmar Meurers
%%%         based on his code originally written June 1997 for the
%%%         communication between ConTroll and CUF, and other code
%%%         by Frederik Fouvry.
%%%
%%% ==================================================================

:- use_module(library(sockets),[hostname_address/2,
                                socket/2,
                                socket_close/1,
                                socket_connect/3]).

%%% --------------------------------------------------------------- %%

start_client:-
   current_prolog_flag(argv,[Hostname,Atom]),
   % two calls to name/2 to turn atom '234' into number 234
   % but leaving file socket names '/tmp/...' unchanged
   name(Atom,Codes),		
   name(PortOrFilename,Codes),
   start_client(Hostname,PortOrFilename).

% start_client(+PortOrFilename)
% e.g., start_client(5002) or start_client('/tmp/fileXYZ').
start_client(PortOrFilename) :-
   start_client(localhost,PortOrFilename).

% start_client(+Hostname,+PortOrFilename)
% e.g., start_client(horatio,5002) or start_client(localhost,'/tmp/fileXYZ')
start_client(Hostname,PortOrFilename) :-
   (  number(PortOrFilename)
   -> start_client_on_port(Hostname,PortOrFilename)
   ;  start_client_on_filesocket(PortOrFilename) % always local, ignores Hostname
   ).
  
stop_client:-
   (  (retract(socket_info(Socket,_)),
       catch(socket_close(Socket),
	     system_error('socket_close/1: ''close'' failed: Bad file number'),
	     print_message(warning,trale(socket(disconnect,failed)))))
   ->  print_message(informational,trale('Client stopped'))
   ;   print_message(warning,trale('No client information found.'))
   ).

%%% --------------------------------------------------------------- %%
% auxiliary predicates:

:- dynamic socket_info/2.
:- volatile socket_info/2.

start_client_on_port(Hostname,Port) :-
   format('Establishing connection to interface on port ~p of ~p.',[Port,Hostname]),
   (  hostname_address(Hostname,IP)
   -> true
   ;  format(user_error,'~N**Error: Cannot determine IP for hostname ~p.~N',[Hostname]),
      abort ),
   socket('AF_INET',Socket),	% This creates the Socket
   catch(socket_connect(Socket,'AF_INET'(IP,Port),Stream),
         system_error('socket_connect/3: ''connect'' failed: Connection refused'),
        (format(user_error,'~N**Error: Connection refused to port ~w of host ~w.~n',[Port,Hostname]),
         fail)
        ),
   asserta(socket_info(Socket,Stream)),
   format('~NConnection established.~n',[]).

start_client_on_filesocket(Filename) :-
   format('Establishing connection to interface on local filesocket ~p.',[Filename]),
   (  name(Filename,[47|_])	% filesocket names start with a /
   -> true
   ;  format(user_error,'~N**Error: File socket name ~p supplied is not an absolute path (i.e., starting with / ).~N',[Filename]),
      abort ),
   socket('AF_UNIX',Socket),	% This creates the Socket
   catch(socket_connect(Socket,'AF_UNIX'(Filename),Stream),
         system_error('socket_connect/3: ''connect'' failed: Connection refused'),
        (format(user_error,'~N**Error: Connection failed to local filesocket ~w.~n',[Filename]),
         fail)
        ),
   asserta(socket_info(Socket,Stream)),
   format('~NConnection established.~n',[]).
