% -*- prolog -*-


% changed the call to capi_printf
% instead wcx_capi_printf is implemented in itsdb.c
% it calls capi_printf. Before calling capi_printf
% the SICStus internal utf8 caracter encoding is converted
% into iso to get the Umlaute reight.

% removed escape_char since it did not load in SICStus 3.9.1
% 04.05.2003

% Stefan Mller <Stefan.Mueller@uni-bremen.de>




% ANSI C - SICStus Prolog interface functions for [incr tsdb()]
% Copyright (C) 2000-2001 Frederik Fouvry
%
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 2 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
%
% Send comments, suggestions, improvements, etc. to Frederik Fouvry
% <fouvry@coli.uni-sb.de>, Computerlinguistik, Universitt des
% Saarlandes, Postfach 15 11 50, D-66041 Saarbrcken, Germany.

:- module(itsdb,[write_field/2,write_field/4,date_time/1,write_table/2]).

:- use_module(library(charsio),[format_to_chars/3,format_to_chars/4]).
:- use_module(library(system),[datime/1]).

:- version('[incr tsdb()] interface for SICStus Prolog applications').

                 
foreign_resource('itsdb',[connect_tsdb,wcx_capi_printf]).

foreign(connect_tsdb,connect_tsdb). % registers capi and calls slave()
foreign(wcx_capi_printf,c,wcx_capi_printf(+chars % char *format
	                                      % ...
									   )).

%%% Load the compiled resource
:- load_foreign_resource(itsdb).

/*
%---------------------------------------------------------------------
% Predicates to call C
% To compile it:
% link_foreign_resource(itsdb,'itsdb.pl',dynamic,['itsdb.c'],[],
%                       ['.../lkb/lib/solaris/libitsdb.so']).

foreign_resource(itsdb, [capi_printf]).
foreign(capi_printf,c,capi_printf(+chars % char *format
	                                      % ...
										   )).
% Load the compiled resource
:- load_foreign_resource(itsdb).

*/

%---------------------------------------------------------------------
% Predicates to format the output
% (uses calls to C) - called in the instantiations of
% user:create_run/5 and user:process_item/7.  Another solution would
% have been for these predicates to return a term with the
% information, which is then passed back up to C, and printed there,
% but that's more complicated ...

% Three types of values: only from processor; from processor with default from itsdb; only from itsdb
% that could be added to tsdb_value and tested for => immediately clear if value will be ignored

write_field(Field,Value):-
	tsdb_value(Field,Type),
	( Type == integer ->
		 format_to_chars('(:~a . ~d) ',[Field,Value],Chars)
	; Type == string -> % quotes within string are not escaped
		 format_to_chars('(:~a . "~w") ',[Field,Value],Chars)
	; Type == date ->
		 format_to_chars('(:~a . "~a") ',[Field,Value],Chars)
	),
	safe_wcx_capi_printf(Chars).
        
write_field(Field,Value,Chars,CharsRest):-
	tsdb_value(Field,Type),
	( Type == integer ->
		 format_to_chars('(:~a . ~d) ',[Field,Value],Chars,CharsRest)
	; Type == string -> % quotes within string are not escaped
		 format_to_chars('(:~a . "~w") ',[Field,Value],Chars,CharsRest)
	; Type == date ->
		 format_to_chars('(:~a . "~a") ',[Field,Value],Chars,CharsRest)
	).


write_table(Table,Records):-
	format_to_chars('~n(:~a ',[Table],Chars1),
	safe_wcx_capi_printf(Chars1),
	write_records(Records),
	format_to_chars(')',[],Chars2),
        safe_wcx_capi_printf(Chars2).


date_time(Date):-
	datime(datime(Y,M,D,H,Min,S)),
	format_to_chars("~|~`0t~d~2+-~`0t~d~3+-~`0t~d~5+ (~`0t~d~3+:~`0t~d~3+:~`0t~d~3+)",
	                [D,M,Y,H,Min,S],Chars),
	atom_codes(Date,Chars).

%---------------------------------------------------------------------
% Auxiliary predicates

write_records([]).
write_records([Record|Rest]):-
	format_to_chars('~n  (~s)',[Record],Chars),
	safe_wcx_capi_printf(Chars),
	write_records(Rest).

% Oes function has a length limitation of 1024, so I have to break
% the string down.
safe_wcx_capi_printf([]).
safe_wcx_capi_printf(Chars) :-
        get_first_n(Chars,0,First,1000,Rest),
        wcx_capi_printf(First),
        safe_wcx_capi_printf(Rest).

get_first_n([],  _,[],_,[]).
get_first_n(Rest,N,[],N,Rest).
get_first_n([H|T],Count,[H|List],N,Rest) :-
  C2 is Count + 1,
  get_first_n(T,C2,List,N,Rest).

%---------------------------------------------------------------------
% Facts to ensure (some form of) correctness
% From the/a relations file - should be generated automatically.
% The tables modified in a run are: parse, run, result and rule.
% daughter and edge are not used anymore.  Their attributes are listed
% here.  Not all of these should be supplied by the interface.
tsdb_value('a-load',integer).
tsdb_value('actives',integer).
tsdb_value('aedges',integer).
tsdb_value('application',string).
tsdb_value('avms',integer).
tsdb_value('comment',string).
tsdb_value('conses',integer).
tsdb_value('copies',integer).
tsdb_value('d-date',date).
tsdb_value('d-end',integer).
tsdb_value('d-key',string).
tsdb_value('d-start',integer).
tsdb_value('d-state',integer).
tsdb_value('d-type',integer).
tsdb_value('d-value',string).
tsdb_value('date',date).
tsdb_value('derivation',string).
tsdb_value('end',date).
tsdb_value('environment',string).
tsdb_value('error',string).
tsdb_value('executed',integer).
tsdb_value('filtered',integer).
tsdb_value('first',integer).
tsdb_value('gcs',integer).
tsdb_value('grammar',string).
tsdb_value('host',string).
tsdb_value('i-id',integer).
tsdb_value('i-load',integer).
tsdb_value('items',integer).
tsdb_value('l-stasks',integer).
tsdb_value('lexicon',integer).
tsdb_value('lrules',integer).
tsdb_value('mrs',string).
tsdb_value('o-application',string).
tsdb_value('o-date',date).
tsdb_value('o-derivation',string).
tsdb_value('o-edges',integer).
tsdb_value('o-gc',integer).
tsdb_value('o-grammar',string).
tsdb_value('o-ignore',string).
tsdb_value('o-mrs',string).
tsdb_value('o-tree',string).
tsdb_value('o-user',string).
tsdb_value('o-wf',integer).
tsdb_value('os',string).
tsdb_value('others',integer).
tsdb_value('p-ctasks',integer).
tsdb_value('p-etasks',integer).
tsdb_value('p-ftasks',integer).
tsdb_value('p-stasks',integer).
tsdb_value('parse-id',integer).
tsdb_value('passives',integer).
tsdb_value('pedges',integer).
tsdb_value('platform',string).
tsdb_value('r-aedges',integer).
tsdb_value('r-ctasks',integer).
tsdb_value('r-etasks',integer).
tsdb_value('r-ftasks',integer).
tsdb_value('r-pedges',integer).
tsdb_value('r-stasks',integer).
tsdb_value('raedges',integer).
tsdb_value('readings',integer).
tsdb_value('result-id',integer).
tsdb_value('rpedges',integer).
tsdb_value('rule',string).
tsdb_value('rules',integer).
tsdb_value('run-comment',string).
tsdb_value('run-id',integer).
tsdb_value('size',integer).
tsdb_value('sorts',integer).
tsdb_value('start',date).
tsdb_value('status',string).
tsdb_value('successes',integer).
tsdb_value('symbols',integer).
tsdb_value('t-active',integer).
tsdb_value('t-author',string).
tsdb_value('t-comment',string).
tsdb_value('t-confidence',integer).
tsdb_value('t-end',date).
tsdb_value('t-start',date).
tsdb_value('t-version',integer).
tsdb_value('tcpu',integer).
tsdb_value('templates',integer).
tsdb_value('tgc',integer).
tsdb_value('time',integer).
tsdb_value('total',integer).
tsdb_value('treal',integer).
tsdb_value('tree',string).
tsdb_value('tsdb',string).
tsdb_value('unifications',integer).
tsdb_value('user',string).
tsdb_value('version',integer).
tsdb_value('words',integer).

%---------------------------------------------------------------------
% Predicates called from C
% It is simpler to write the error term in Prolog than in C

create_run(TSInstanceName,RunID,Comment,Interactivep,Command):-
	user:catch(create_run(TSInstanceName,RunID,Comment,Interactivep,
	                       Command),
	           Term,
			     (write(user_error,Term),nl(user_error),fail)).
process_item(ItemID,ItemInput,ParseID,MaxEdges,Exhaustivep,
	          Derivationp,Interactivep):-
	user:catch(process_item(ItemID,ItemInput,ParseID,MaxEdges,
	                         Exhaustivep,Derivationp,Interactivep),
	           Term,
			     (write(user_error,Term),nl(user_error),
					itsdb:write_field('error','Caught Prolog exception'),
					fail)).
complete_run(RunID,AppSpecific):-
	user:catch(complete_run(RunID,AppSpecific),
	           Term,
			     (write(user_error,Term),nl(user_error),fail)).
reconstruct_item(ItemID):-
	user:catch(reconstruct_item(ItemID),
	           Term,
			     (write(user_error,Term),nl(user_error),fail)).
